@@ -143,6 +143,8 @@ static void compute_allocation_strategy(MVMThreadContext *tc, MVMObject *repr_in
             /* Fetch its type; see if it's some kind of unboxed type. */
             MVMObject *attr  = MVM_repr_at_pos_o(tc, flat_list, i);
             MVMObject *type  = MVM_repr_at_key_o(tc, attr, tc->instance->str_consts.type);
+            MVMObject *inlined_val = MVM_repr_at_key_o(tc, attr, tc->instance->str_consts.inlined);
+            MVMint64 inlined = !MVM_is_null(tc, inlined_val) && MVM_repr_get_int(tc, inlined_val);
             MVMint32   bits  = sizeof(void *) * 8;
             MVMint32   align = ALIGNOF(void *);
             if (!MVM_is_null(tc, type)) {
@@ -160,11 +162,6 @@ static void compute_allocation_strategy(MVMThreadContext *tc, MVMObject *repr_in
                     bits = spec->bits;
                     align = spec->align;
 
-                    if (bits % 8) {
-                         MVM_exception_throw_adhoc(tc,
-                            "CStruct only supports native types that are a multiple of 8 bits wide (was passed: %"PRId32")", bits);
-                    }
-
                     repr_data->attribute_locations[i] = (bits << MVM_CSTRUCT_ATTR_SHIFT) | MVM_CSTRUCT_ATTR_IN_STRUCT;
                     repr_data->flattened_stables[i] = STABLE(type);
                     if (REPR(type)->initialize) {
@@ -191,6 +188,24 @@ static void compute_allocation_strategy(MVMThreadContext *tc, MVMObject *repr_in
                     repr_data->num_child_objs++;
                     repr_data->attribute_locations[i] = (cur_obj_attr++ << MVM_CSTRUCT_ATTR_SHIFT) | MVM_CSTRUCT_ATTR_CSTRUCT;
                     repr_data->member_types[i] = type;
+                    if (inlined) {
+                        MVMCStructREPRData *cstruct_repr_data = (MVMCStructREPRData *)STABLE(type)->REPR_data;
+                        bits                                  = cstruct_repr_data->struct_size * 8;
+                        align                                 = cstruct_repr_data->struct_size;
+                        repr_data->attribute_locations[i]    |= MVM_CSTRUCT_ATTR_INLINED;
+                    }
+                }
+                else if (type_id == MVM_REPR_ID_MVMCUnion) {
+                    /* It's a CUnion. */
+                    repr_data->num_child_objs++;
+                    repr_data->attribute_locations[i] = (cur_obj_attr++ << MVM_CSTRUCT_ATTR_SHIFT) | MVM_CSTRUCT_ATTR_CUNION;
+                    repr_data->member_types[i] = type;
+                    if (inlined) {
+                        MVMCUnionREPRData *cunion_repr_data = (MVMCUnionREPRData *)STABLE(type)->REPR_data;
+                        bits                                = cunion_repr_data->struct_size * 8;
+                        align                               = cunion_repr_data->struct_size;
+                        repr_data->attribute_locations[i]  |= MVM_CSTRUCT_ATTR_INLINED;
+                    }
                 }
                 else if (type_id == MVM_REPR_ID_MVMCPointer) {
                     /* It's a CPointer. */
@@ -208,6 +223,11 @@ static void compute_allocation_strategy(MVMThreadContext *tc, MVMObject *repr_in
                     "CStruct representation requires the types of all attributes to be specified");
             }
 
+            if (bits % 8) {
+                 MVM_exception_throw_adhoc(tc,
+                    "CStruct only supports native types that are a multiple of 8 bits wide (was passed: %"PRId32")", bits);
+            }
+
             /* Do allocation. */
             /* C structure needs careful alignment. If cur_size is not aligned
              * to align bytes (cur_size % align), make sure it is before we
@@ -225,6 +245,8 @@ static void compute_allocation_strategy(MVMThreadContext *tc, MVMObject *repr_in
 
         /* Finally, put computed allocation size in place; it's body size plus
          * header size. Also number of markables and sentinels. */
+        if (multiple_of > sizeof(void *))
+            multiple_of = sizeof(void *);
         repr_data->struct_size = ceil((double)cur_size / (double)multiple_of) * multiple_of;
         if (repr_data->initialize_slots)
             repr_data->initialize_slots[cur_init_slot] = -1;
@@ -360,7 +382,18 @@ static void get_attribute(MVMThreadContext *tc, MVMSTable *st, MVMObject *root,
                             obj = MVM_nativecall_make_carray(tc, typeobj, cobj);
                         }
                         else if(type == MVM_CSTRUCT_ATTR_CSTRUCT) {
-                            obj = MVM_nativecall_make_cstruct(tc, typeobj, cobj);
+                            if (repr_data->attribute_locations[slot] & MVM_CSTRUCT_ATTR_INLINED)
+                                obj = MVM_nativecall_make_cstruct(tc, typeobj,
+                                    (char *)body->cstruct + repr_data->struct_offsets[slot]);
+                            else
+                                obj = MVM_nativecall_make_cstruct(tc, typeobj, cobj);
+                        }
+                        else if(type == MVM_CSTRUCT_ATTR_CUNION) {
+                            if (repr_data->attribute_locations[slot] & MVM_CSTRUCT_ATTR_INLINED)
+                                obj = MVM_nativecall_make_cunion(tc, typeobj,
+                                    (char *)body->cstruct + repr_data->struct_offsets[slot]);
+                            else
+                                obj = MVM_nativecall_make_cunion(tc, typeobj, cobj);
                         }
                         else if(type == MVM_CSTRUCT_ATTR_CPTR) {
                             obj = MVM_nativecall_make_cpointer(tc, typeobj, cobj);
@@ -459,6 +492,12 @@ static void bind_attribute(MVMThreadContext *tc, MVMSTable *st, MVMObject *root,
                                 "Can only store CStruct attribute in CStruct slot in CStruct");
                         cobj = ((MVMCStruct *)value)->body.cstruct;
                     }
+                    else if (type == MVM_CSTRUCT_ATTR_CUNION) {
+                        if (REPR(value)->ID != MVM_REPR_ID_MVMCUnion)
+                            MVM_exception_throw_adhoc(tc,
+                                "Can only store CUnion attribute in CUnion slot in CStruct");
+                        cobj = ((MVMCUnion *)value)->body.cunion;
+                    }
                     else if (type == MVM_CSTRUCT_ATTR_CPTR) {
                         if (REPR(value)->ID != MVM_REPR_ID_MVMCPointer)
                             MVM_exception_throw_adhoc(tc,
