@@ -14,11 +14,7 @@ typedef struct {
 } BoolMethReturnData;
 
 MVMint64 MVM_coerce_istrue_s(MVMThreadContext *tc, MVMString *str) {
-    return str == NULL ||
-           !IS_CONCRETE(str) ||
-           MVM_string_graphs(tc, str) == 0 ||
-           (MVM_string_graphs(tc, str) == 1 && MVM_string_get_grapheme_at_nocheck(tc, str, 0) == 48)
-           ? 0 : 1;
+    return str == NULL || !IS_CONCRETE(str) || MVM_string_graphs(tc, str) == 0 ? 0 : 1;
 }
 
 /* Tries to do the boolification. It may be that a method call is needed. In
@@ -75,10 +71,7 @@ void MVM_coerce_istrue(MVMThreadContext *tc, MVMObject *obj, MVMRegister *res_re
             case MVM_BOOL_MODE_UNBOX_NUM:
                 result = !IS_CONCRETE(obj) || REPR(obj)->box_funcs.get_num(tc, STABLE(obj), obj, OBJECT_BODY(obj)) == 0.0 ? 0 : 1;
                 break;
-            case MVM_BOOL_MODE_UNBOX_STR_NOT_EMPTY:
-                result = !IS_CONCRETE(obj) || MVM_string_graphs(tc, REPR(obj)->box_funcs.get_str(tc, STABLE(obj), obj, OBJECT_BODY(obj))) == 0 ? 0 : 1;
-                break;
-            case MVM_BOOL_MODE_UNBOX_STR_NOT_EMPTY_OR_ZERO: {
+            case MVM_BOOL_MODE_UNBOX_STR_NOT_EMPTY: {
                 MVMString *str;
                 if (!IS_CONCRETE(obj)) {
                     result = 0;
@@ -88,6 +81,19 @@ void MVM_coerce_istrue(MVMThreadContext *tc, MVMObject *obj, MVMRegister *res_re
                 result = MVM_coerce_istrue_s(tc, str);
                 break;
             }
+            case MVM_BOOL_MODE_UNBOX_STR_NOT_EMPTY_OR_ZERO: {
+                MVMString *str;
+                if (!IS_CONCRETE(obj)) {
+                    result = 0;
+                    break;
+                }
+                str = REPR(obj)->box_funcs.get_str(tc, STABLE(obj), obj, OBJECT_BODY(obj));
+                result = str == NULL ||
+                        !IS_CONCRETE(str) ||
+                        (MVM_string_graphs(tc, str) == 1 && MVM_string_get_grapheme_at_nocheck(tc, str, 0) == 48)
+                        ? 0 : 1;
+                break;
+            }
             case MVM_BOOL_MODE_NOT_TYPE_OBJECT:
                 result = !IS_CONCRETE(obj) ? 0 : 1;
                 break;
@@ -342,7 +348,7 @@ MVMObject * MVM_radix(MVMThreadContext *tc, MVMint64 radix, MVMString *str, MVMi
     MVMint64   ch;
 
     if (radix > 36) {
-        MVM_exception_throw_adhoc(tc, "Cannot convert radix of %d (max 36)", radix);
+        MVM_exception_throw_adhoc(tc, "Cannot convert radix of %"PRId64" (max 36)", radix);
     }
 
     ch = (offset < chars) ? MVM_string_get_grapheme_at_nocheck(tc, str, offset) : 0;
