@@ -24,6 +24,7 @@ MVMFixedSizeAlloc * MVM_fixed_size_create(MVMThreadContext *tc) {
     if ((init_stat = uv_mutex_init(&(al->complex_alloc_mutex))) < 0)
         MVM_exception_throw_adhoc(tc, "Failed to initialize mutex: %s",
             uv_strerror(init_stat));
+    al->freelist_spin = 0;
     return al;
 }
 
@@ -77,7 +78,7 @@ static void * alloc_slow_path(MVMThreadContext *tc, MVMFixedSizeAlloc *al, MVMui
     void *result;
 
     /* Lock, unless single-threaded. */
-    MVMint32 lock = tc->instance->next_user_thread_id != 2;
+    MVMint32 lock = MVM_instance_have_user_threads(tc);
     if (lock)
         uv_mutex_lock(&(al->complex_alloc_mutex));
 
@@ -110,13 +111,23 @@ void * MVM_fixed_size_alloc(MVMThreadContext *tc, MVMFixedSizeAlloc *al, size_t
         /* Try and take from the free list (fast path). */
         MVMFixedSizeAllocSizeClass     *bin_ptr = &(al->size_classes[bin]);
         MVMFixedSizeAllocFreeListEntry *fle;
-        if (tc->instance->next_user_thread_id != 2) {
-            /* Multi-threaded; race for it. */
+        if (MVM_instance_have_user_threads(tc)) {
+            /* Multi-threaded; take a lock. Note that the lock is needed in
+             * addition to the atomic operations: the atomics allow us to add
+             * to the free list in a lock-free way, and the lock allows us to
+             * avoid the ABA issue we'd have with only the atomics. */
+            while (!MVM_trycas(&(al->freelist_spin), 0, 1)) {
+                MVMint32 i = 0;
+                while (i < 1024)
+                    i++;
+            }
             do {
                 fle = bin_ptr->free_list;
                 if (!fle)
                     break;
             } while (!MVM_trycas(&(bin_ptr->free_list), fle, fle->next));
+            MVM_barrier();
+            al->freelist_spin = 0;
         }
         else {
             /* Single-threaded; just take it. */
@@ -158,7 +169,7 @@ void MVM_fixed_size_free(MVMThreadContext *tc, MVMFixedSizeAlloc *al, size_t byt
         MVMFixedSizeAllocSizeClass     *bin_ptr = &(al->size_classes[bin]);
         MVMFixedSizeAllocFreeListEntry *to_add  = (MVMFixedSizeAllocFreeListEntry *)to_free;
         MVMFixedSizeAllocFreeListEntry *orig;
-        if (tc->instance->next_user_thread_id != 2) {
+        if (MVM_instance_have_user_threads(tc)) {
             /* Multi-threaded; race to add it. */
             do {
                 orig = bin_ptr->free_list;
@@ -195,7 +206,7 @@ void MVM_fixed_size_free_at_safepoint(MVMThreadContext *tc, MVMFixedSizeAlloc *a
         MVMFixedSizeAllocSizeClass     *bin_ptr = &(al->size_classes[bin]);
         MVMFixedSizeAllocFreeListEntry *to_add  = (MVMFixedSizeAllocFreeListEntry *)to_free;
         MVMFixedSizeAllocFreeListEntry *orig;
-        if (tc->instance->next_user_thread_id != 2) {
+        if (MVM_instance_have_user_threads(tc)) {
             /* Multi-threaded; race to add it to the "free at next safe point"
              * list. */
             MVM_panic(1, "MVM_fixed_size_free_at_safepoint not yet fully implemented");
@@ -209,7 +220,7 @@ void MVM_fixed_size_free_at_safepoint(MVMThreadContext *tc, MVMFixedSizeAlloc *a
     }
     else {
         /* Was malloc'd due to being oversize. */
-        if (tc->instance->next_user_thread_id != 2) {
+        if (MVM_instance_have_user_threads(tc)) {
             /* Multi-threaded; race to add it to the "free at next safe point"
              * list. */
             MVM_panic(1, "MVM_fixed_size_free_at_safepoint not yet fully implemented");
