@@ -51,6 +51,7 @@ struct MVMStringConsts {
     MVMString *free_str;
     MVMString *callback_args;
     MVMString *encoding;
+    MVMString *inlined;
     MVMString *repr;
     MVMString *anon;
     MVMString *P6opaque;
@@ -160,8 +161,10 @@ struct MVMInstance {
     /* int -> str cache */
     MVMString **int_to_str_cache;
 
-    /* Specialization installation mutex (global, as it's low contention, so
-     * no real motivation to have it more fine-grained at present). */
+    /* Multi-dispatch cache and specialization installation mutexes
+     * (global, as the additions are quite low contention, so no
+     * real motivation to have it more fine-grained at present). */
+    uv_mutex_t mutex_multi_cache_add;
     uv_mutex_t mutex_spesh_install;
 
     /* Log file for specializations, if we're to log them. */
@@ -261,9 +264,9 @@ struct MVMInstance {
     MVMIntConstCache    *int_const_cache;
     uv_mutex_t mutex_int_const_cache;
 
-    /* Atomically-incremented counter of newly invoked frames,
-     * so each can obtain an index into each threadcontext's pool table */
-    AO_t num_frame_pools;
+    /* Atomically-incremented counter of newly invoked frames, used for
+     * lexotic caching. */
+    AO_t num_frames_run;
 
     /* Hash of compiler objects keyed by name */
     MVMObject          *compiler_registry;
@@ -330,4 +333,13 @@ struct MVMInstance {
 
     /* Whether profiling is turned on or not. */
     MVMuint32 profiling;
+
+    /* Cached backend config hash. */
+    MVMObject *cached_backend_config;
 };
+
+/* Returns a true value if we have created user threads (and so are running a
+ * multi-threaded application). */
+MVM_STATIC_INLINE MVMint32 MVM_instance_have_user_threads(MVMThreadContext *tc) {
+    return tc->instance->next_user_thread_id != 2;
+}
