@@ -91,7 +91,7 @@ MVMInstance * MVM_vm_create_instance(void) {
                 instance->main_thread, STABLE(instance->boot_types.BOOTThread))));
     instance->threads->body.stage = MVM_thread_stage_started;
     instance->threads->body.tc = instance->main_thread;
-    instance->threads->body.thread_id = MVM_platform_thread_id();
+    instance->threads->body.native_thread_id = MVM_platform_thread_id();
 
     /* Create compiler registry */
     instance->compiler_registry = MVM_repr_alloc_init(instance->main_thread, instance->boot_types.BOOTHash);
@@ -119,6 +119,9 @@ MVMInstance * MVM_vm_create_instance(void) {
      * them, so that spesh may end up optimizing more "internal" stuff. */
     MVM_callsite_initialize_common(instance->main_thread);
 
+    /* Multi-cache additions mutex. */
+    init_mutex(instance->mutex_multi_cache_add, "multi-cache addition");
+
     /* Mutex for spesh installations, and check if we've a file we
      * should log specializations to. */
     init_mutex(instance->mutex_spesh_install, "spesh installations");
@@ -136,6 +139,7 @@ MVMInstance * MVM_vm_create_instance(void) {
             instance->spesh_osr_enabled = 1;
     }
 
+    /* JIT environment/logging setup. */
     jit_disable = getenv("MVM_JIT_DISABLE");
     if (!jit_disable || strlen(jit_disable) == 0)
         instance->jit_enabled = 1;
@@ -298,6 +302,9 @@ void MVM_vm_destroy_instance(MVMInstance *instance) {
     /* Clean up Hash of hashes of symbol tables per hll. */
     uv_mutex_destroy(&instance->mutex_hll_syms);
 
+    /* Clean up multi cache addition mutex. */
+    uv_mutex_destroy(&instance->mutex_multi_cache_add);
+
     /* Clean up spesh install mutex and close any log. */
     uv_mutex_destroy(&instance->mutex_spesh_install);
     if (instance->spesh_log_fh)
