@@ -51,12 +51,45 @@ static void dump_bb(MVMThreadContext *tc, DumpStr *ds, MVMSpeshGraph *g, MVMSpes
     /* Heading. */
     appendf(ds, "  BB %d (%p):\n", bb->idx, bb);
 
+    if (bb->inlined) {
+        append(ds, "    Inlined\n");
+    }
+
+    {
+        /* Also, we have a line number */
+        MVMBytecodeAnnotation *bbba = MVM_bytecode_resolve_annotation(tc, &g->sf->body, bb->initial_pc);
+        MVMuint32 line_number;
+        if (bbba) {
+            line_number = bbba->line_number;
+            MVM_free(bbba);
+        } else {
+            line_number = -1;
+        }
+        appendf(ds, "    line: %d (pc %d)\n", line_number, bb->initial_pc);
+    }
+
     /* Instructions. */
     append(ds, "    Instructions:\n");
     cur_ins = bb->first_ins;
     while (cur_ins) {
         MVMSpeshAnn *ann = cur_ins->annotations;
+        MVMuint32 line_number;
+
         while (ann) {
+            /* These four annotations carry a deopt index that we can find a
+             * corresponding line number for */
+            if (ann->type == MVM_SPESH_ANN_DEOPT_ONE_INS
+                || ann->type == MVM_SPESH_ANN_DEOPT_ALL_INS
+                || ann->type == MVM_SPESH_ANN_DEOPT_INLINE
+                || ann->type == MVM_SPESH_ANN_DEOPT_OSR) {
+                MVMBytecodeAnnotation *ba = MVM_bytecode_resolve_annotation(tc, &g->sf->body, g->deopt_addrs[2 * ann->data.deopt_idx]);
+                if (ba) {
+                    line_number = ba->line_number;
+                    MVM_free(ba);
+                } else {
+                    line_number = -1;
+                }
+            }
             switch (ann->type) {
                 case MVM_SPESH_ANN_FH_START:
                     appendf(ds, "      [Annotation: FH Start (%d)]\n",
@@ -71,12 +104,12 @@ static void dump_bb(MVMThreadContext *tc, DumpStr *ds, MVMSpeshGraph *g, MVMSpes
                         ann->data.frame_handler_index);
                     break;
                 case MVM_SPESH_ANN_DEOPT_ONE_INS:
-                    appendf(ds, "      [Annotation: INS Deopt One (idx %d -> pc %d)]\n",
-                        ann->data.deopt_idx, g->deopt_addrs[2 * ann->data.deopt_idx]);
+                    appendf(ds, "      [Annotation: INS Deopt One (idx %d -> pc %d; line %d)]\n",
+                        ann->data.deopt_idx, g->deopt_addrs[2 * ann->data.deopt_idx], line_number);
                     break;
                 case MVM_SPESH_ANN_DEOPT_ALL_INS:
-                    appendf(ds, "      [Annotation: INS Deopt All (idx %d -> pc %d)]\n",
-                        ann->data.deopt_idx, g->deopt_addrs[2 * ann->data.deopt_idx]);
+                    appendf(ds, "      [Annotation: INS Deopt All (idx %d -> pc %d; line %d)]\n",
+                        ann->data.deopt_idx, g->deopt_addrs[2 * ann->data.deopt_idx], line_number);
                     break;
                 case MVM_SPESH_ANN_INLINE_START:
                     appendf(ds, "      [Annotation: Inline Start (%d)]\n",
@@ -87,12 +120,12 @@ static void dump_bb(MVMThreadContext *tc, DumpStr *ds, MVMSpeshGraph *g, MVMSpes
                         ann->data.inline_idx);
                     break;
                 case MVM_SPESH_ANN_DEOPT_INLINE:
-                    appendf(ds, "      [Annotation: INS Deopt Inline (idx %d -> pc %d)]\n",
-                        ann->data.deopt_idx, g->deopt_addrs[2 * ann->data.deopt_idx]);
+                    appendf(ds, "      [Annotation: INS Deopt Inline (idx %d -> pc %d; line %d)]\n",
+                        ann->data.deopt_idx, g->deopt_addrs[2 * ann->data.deopt_idx], line_number);
                     break;
                 case MVM_SPESH_ANN_DEOPT_OSR:
-                    appendf(ds, "      [Annotation: INS Deopt OSR (idx %d -> pc %d)]\n",
-                        ann->data.deopt_idx, g->deopt_addrs[2 * ann->data.deopt_idx]);
+                    appendf(ds, "      [Annotation: INS Deopt OSR (idx %d -> pc %d); line %d]\n",
+                        ann->data.deopt_idx, g->deopt_addrs[2 * ann->data.deopt_idx], line_number);
                     break;
                 default:
                     appendf(ds, "      [Annotation: %d (unknown)]\n", ann->type);
@@ -184,6 +217,23 @@ static void dump_bb(MVMThreadContext *tc, DumpStr *ds, MVMSpeshGraph *g, MVMSpes
                         case MVM_operand_spesh_slot:
                             appendf(ds, "sslot(%"PRId16")", cur_ins->operands[i].lit_i16);
                             break;
+                        case MVM_operand_coderef: {
+                            MVMCodeBody *body = &((MVMCode*)g->sf->body.cu->body.coderefs[cur_ins->operands[i].coderef_idx])->body;
+                            MVMBytecodeAnnotation *ann = MVM_bytecode_resolve_annotation(tc, &body->sf->body, 0);
+
+                            append(ds, "coderef(");
+
+                            if (ann) {
+                                char *filestr = MVM_string_utf8_encode_C_string(tc, g->sf->body.cu->body.strings[ann->filename_string_heap_index]);
+                                appendf(ds, "%s:%d%s)", filestr, ann->line_number, body->outer ? " (closure)" : "");
+                                MVM_free(filestr);
+                            } else {
+                                append(ds, "??\?)");
+                            }
+
+                            MVM_free(ann);
+                            break;
+                        }
                         default:
                             append(ds, "<nyi(lit)>");
                         }
@@ -220,7 +270,44 @@ static void dump_facts(MVMThreadContext *tc, DumpStr *ds, MVMSpeshGraph *g) {
         for (j = 0; j < num_facts; j++) {
             MVMint32 usages = g->facts[i][j].usages;
             MVMint32 flags  = g->facts[i][j].flags;
-            appendf(ds, "    r%d(%d): usages=%d, flags=%d\n", i, j, usages, flags);
+            appendf(ds, "    r%d(%d): usages=%d, flags=%d", i, j, usages, flags);
+            if (flags & 1) {
+                append(ds, " KnTyp");
+            }
+            if (flags & 2) {
+                append(ds, " KnVal");
+            }
+            if (flags & 4) {
+                append(ds, " Dcntd");
+            }
+            if (flags & 8) {
+                append(ds, " Concr");
+            }
+            if (flags & 16) {
+                append(ds, " TyObj");
+            }
+            if (flags & 32) {
+                append(ds, " KnDcT");
+            }
+            if (flags & 64) {
+                append(ds, " DCncr");
+            }
+            if (flags & 128) {
+                append(ds, " DcTyO");
+            }
+            if (flags & 256) {
+                append(ds, " LogGd");
+            }
+            if (flags & 512) {
+                append(ds, " HashI");
+            }
+            if (flags & 1024) {
+                append(ds, " ArrIt");
+            }
+            if (flags & 2048) {
+                append(ds, " KBxSr");
+            }
+            append(ds, "\n");
         }
     }
 }
@@ -230,7 +317,7 @@ static void dump_callsite(MVMThreadContext *tc, DumpStr *ds, MVMSpeshGraph *g) {
     appendf(ds, "Callsite %p (%d args, %d pos)\n", g->cs, g->cs->arg_count, g->cs->num_pos);
     for (i = 0; i < (g->cs->arg_count - g->cs->num_pos) / 2; i++) {
         if (g->cs->arg_names[i]) {
-            char * argname_utf8 = MVM_string_utf8_encode(tc, g->cs->arg_names[i], NULL);
+            char * argname_utf8 = MVM_string_utf8_encode_C_string(tc, g->cs->arg_names[i]);
             appendf(ds, "  - %s\n", argname_utf8);
             MVM_free(argname_utf8);
         }
@@ -249,7 +336,7 @@ static void dump_fileinfo(MVMThreadContext *tc, DumpStr *ds, MVMSpeshGraph *g) {
         filename = cu->body.strings[str_idx];
     }
     if (filename)
-        filename_utf8 = MVM_string_utf8_encode(tc, filename, NULL);
+        filename_utf8 = MVM_string_utf8_encode_C_string(tc, filename);
     appendf(ds, "%s:%d", filename_utf8, line_nr);
     if (filename)
         MVM_free(filename_utf8);
