@@ -127,6 +127,69 @@ static MVMuint8 windows1252_cp_to_char(MVMint32 codepoint) {
 
 }
 
+/* Decodes using a decodestream. Decodes as far as it can with the input
+ * buffers, or until a stopper is reached. */
+void MVM_string_windows1252_decodestream(MVMThreadContext *tc, MVMDecodeStream *ds,
+                                    MVMint32 *stopper_chars, MVMint32 *stopper_sep) {
+    MVMint32 count = 0, total = 0;
+    MVMint32 bufsize;
+    MVMGrapheme32 *buffer;
+    MVMDecodeStreamBytes *cur_bytes;
+    MVMDecodeStreamBytes *last_accept_bytes = ds->bytes_head;
+    MVMint32 last_accept_pos;
+
+    /* If there's no buffers, we're done. */
+    if (!ds->bytes_head)
+        return;
+    last_accept_pos = ds->bytes_head_pos;
+
+    /* If we're asked for zero chars, also done. */
+    if (stopper_chars && *stopper_chars == 0)
+        return;
+
+    /* Take length of head buffer as initial guess. */
+    bufsize = ds->bytes_head->length;
+    buffer = MVM_malloc(bufsize * sizeof(MVMGrapheme32));
+
+    /* Decode each of the buffers. */
+    cur_bytes = ds->bytes_head;
+    while (cur_bytes) {
+        /* Process this buffer. */
+        MVMint32  pos = cur_bytes == ds->bytes_head ? ds->bytes_head_pos : 0;
+        unsigned char *bytes = (unsigned char *)cur_bytes->bytes;
+        while (pos < cur_bytes->length) {
+            MVMCodepoint codepoint = WINDOWS1252_CHAR_TO_CP(bytes[pos++]);
+            if (count == bufsize) {
+                /* We filled the buffer. Attach this one to the buffers
+                 * linked list, and continue with a new one. */
+                MVM_string_decodestream_add_chars(tc, ds, buffer, bufsize);
+                buffer = MVM_malloc(bufsize * sizeof(MVMGrapheme32));
+                count = 0;
+            }
+            buffer[count++] = codepoint; /* XXX NFG needs this to change. */
+            last_accept_bytes = cur_bytes;
+            last_accept_pos = pos;
+            total++;
+            if (stopper_chars && *stopper_chars == total)
+                goto done;
+            if (stopper_sep && *stopper_sep == codepoint)
+                goto done;
+        }
+        cur_bytes = cur_bytes->next;
+    }
+  done:
+
+    /* Attach what we successfully parsed as a result buffer, and trim away
+     * what we chewed through. */
+    if (count) {
+        MVM_string_decodestream_add_chars(tc, ds, buffer, count);
+    }
+    else {
+	MVM_free(buffer);
+    }
+    MVM_string_decodestream_discard_to(tc, ds, last_accept_bytes, last_accept_pos);
+}
+
 /* Decodes the specified number of bytes of windows1252 into an NFG string,
  * creating a result of the specified type. The type must have the MVMString
  * REPR. */
